/*
 * Copyright (c) 2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * Test program for ABR.
 *
 * This program let's you play with ABR. You need two things for this:
 * a card that allows you to send/receive raw cells (hatm only at the moment).
 * And a card that you want to test that supports ABR. The latter not
 * necessarily in a FreeBSD box.
 *
 * You need the following setup:
 *
 *                              switch
 * +--------------+           +--------+  vci+1    +----------+
 * | box with ABR |  vci      |   +---->-----------> FreeBSD  |
 * |  adapter     |---------------+    |  vci      | with hatm|
 * |              |           |   +----<-----------<          |
 * +--------------+           +--------+           +----------+
 *
 * That is you must split your VC on the switch so that the FreeBSD box
 * receives on vci and transmits on vci+1 (the vcis on the left and the
 * reight sides need not be to be equal. The reason for this is that the
 * ForeRunnerHE seems to process received RM cells and reflects them even
 * if told not to do so. By splitting the VC the reflected cells go into
 * the bit bucket.
 *
 * Start the program.
 *
 * Every time when the remote ABR adapter changes its ER the program will
 * you tell this. You may toggle influence fields of the reflected RM cells:
 * CI, NI, BECM and ER. You may enter the following to the program:
 *
 *  s	- toggle stop. When stopped RM cells are not reflected. The card
 *	  should reduce its ACR with the CDF down to MCR.
 *  n	- toggle the NI (no increase) flag. When set the card should not
 *	  increase its ACR.
 *  c	- toggle the CI (congestion) flag. When set the card should reduce
 *	  its ACR with RDF to MCR.
 *  NNN	- a number. Set the reflected ER to this value. The card should reduce
 *	  its rate to this value  if it is currently at a higher rate but not
 *	  below MCR. It should increase if it is at a lower rate, but not when
 *	  NI is set and not higher than PCR. It should not change its rate
 *	  when CI is set.
 *  x	- reset the above number. The reflected ER is the received ER.
 *  g	- toggle the BECM flag.
 *  \n	- print settings
 *  b	- send an unsolicited BRM cell with current settings.
 *
 * $Id: atmabr.c,v 1.3 2003/07/14 15:14:44 catscvs Exp $
 */
#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <err.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <math.h>
#include <limits.h>

#include <netgraph.h>
#include <netgraph/atm/ng_atm.h>

#include <net/if_atm.h>

static struct ngm_atm_cpcs_init tparam = {
	vci:	66,
	rmtu:	10000,
	tmtu:	10000,
	aal:	ATMIO_AAL_RAW,
	traffic:ATMIO_TRAFFIC_UBR,
};
static struct ngm_atm_cpcs_init rparam = {
	vci:	67,
	rmtu:	10000,
	tmtu:	10000,
	aal:	ATMIO_AAL_RAW,
	traffic:ATMIO_TRAFFIC_UBR,
};

#define	RXHOOK	"rx"
#define	TXHOOK	"tx"

static int verbose;
static int csock, dsock;
static int kq;

struct count {
	uint64_t	total;
	uint64_t	last;
};

static struct count dcells;	/* data cells */
static struct count frm0cells;	/* RM CLP0 cells received */
static struct count frm1cells;	/* RM CLP1 cells received */
static struct count brm0cells;	/* BRM CLP0 cells received */
static struct count brm1cells;	/* BRM CLP1 cells received */
static struct count ucells;	/* all other cells */

static uint64_t t_start;
static uint64_t t_last;

static const char usgtxt[] = "\
ATM ABR tester.\n\
usage: atmabr [-hsv] [-c [vpi.[vci]] [-d dev]\n\
  -c	specify channel to use\n\
  -d	use specified device instead of hatm0\n\
  -h	print this help\n\
  -s	start in stopped state\n\
  -v	be verbose\n\
";

static uint16_t crc10tab[256];
static double	last_er = 0;
static double	new_er = -1;
static u_int	ni, ci, ng;
static int	stop;

static void
gen_crc10_tab(void)
{
	int i, j;
	uint16_t crc;

	for (i = 0; i < 256; i++) {
		crc = i << 2;
		for (j = 0; j < 8; j++)
			if ((crc <<= 1) & 0x400)
				crc ^= 0x633;
		crc10tab[i] = crc;
	}
}

static uint16_t
gen_crc10(uint16_t crc, const u_char *ptr, u_int bytes)
{

	while (bytes--)
		crc = ((crc << 8) & 0x3ff) ^
		    crc10tab[(crc >> 2) & 0xff] ^
		    *ptr++;
	return (crc);
}

static double
atmf2cr(u_int atmf)
{
	return ((1 << ((atmf >> 9) & 0x1f)) * (512 + (atmf & 0x1ff)) / 512.0
	  * (atmf >> 14));
}

static u_int
cps2atmf(double cps)
{
	u_int s, id;

	if (cps < 1.0)
		return (0);

	s = 0;
	while (cps >= 2.0) {
		s++;
		cps /= 2;
	}
	cps = rint(512 * cps);
	id = (u_int)cps;

	return ((1 << 14) | (s << 9) | (id & 0x1ff));
}

/*
 * Print statistics
 */
static void
do_info(struct count *c, uint64_t t)
{
	printf("%llu/%.3f (%llu/%.3f)", (unsigned long long)c->total,
	    1000000.0 * c->total / (t - t_start),
	    (unsigned long long)(c->total - c->last),
	    1000000.0 * (c->total - c->last) / (t - t_last));
	c->last = c->total;
}

static void
print_info(void)
{
	struct timeval tv;
	uint64_t t;

	gettimeofday(&tv, NULL);
	t = tv.tv_sec * 1000000ULL + tv.tv_usec;

	printf("data cells: ");
	do_info(&dcells, t);
	printf(";  unknown cells: ");
	do_info(&ucells, t);
	printf("\n");

	printf("FRM cells clp0: ");
	do_info(&frm0cells, t);
	printf(";  clp1: ");
	do_info(&frm1cells, t);
	printf("\n");

	printf("BRM cells clp0: ");
	do_info(&brm0cells, t);
	printf(";  clp1: ");
	do_info(&brm1cells, t);
	printf("\n");

	t_last = t;
}

/*
 * Make a cell header from our well known VPI/VCI
 */
static void
cell_mkheader(u_char *cell, u_int pti)
{
	cell[0] = tparam.vpi >> 4;
	cell[1] = (tparam.vpi << 4) | (tparam.vci >> 12);
	cell[2] = tparam.vci << 4;
	cell[3] = (tparam.vci << 4) | (pti << 1);
}

/*
 * Send a cell
 */
static void
cell_send(u_char *cell)
{
	int ret;

	/* send it */
	ret = NgSendData(dsock, TXHOOK, cell, 52);
	if (ret == -1)
		err(1, "send");
}

/*
 * Timer expired - make a forward RM cell
 */
static void
process_timer(void)
{
	u_char cell[52];
	uint16_t crc;

	bzero(cell, sizeof(cell));
	cell_mkheader(cell, 6);

	cell[4] = 1;

	crc = gen_crc10(0, cell + 4, 48);
	cell[50] |= crc >> 8;
	cell[51] = crc;

	cell_send(cell);
}

/*
 * Dump a resource management cell
 */
static void
rm_dump(const u_char *cell)
{
	u_int i;

	if (verbose > 1) {
		for (i = 0; i < 52; i++)
			printf("%02x ", cell[i]);
		printf("\n\n");
	}
	printf("RM: %s%s%s%s ecr=%f ccr=%f mcr=%f\n",
	    (cell[5] & 0x80) ? "B" : "F", (cell[5] & 0x40) ? " BECN" : "",
	    (cell[5] & 0x20) ? " CI" : "", (cell[5] & 0x10) ? " NI" : "",
	    atmf2cr((cell[6] << 8) | cell[7]),
	    atmf2cr((cell[8] << 8) | cell[9]),
	    atmf2cr((cell[10] << 8) | cell[11]));
}

/*
 * Make a resource management cell (without the header)
 */
static void
rm_makecell(u_char *cell, u_char flags, double er, double ccr, double mcr)
{
	u_int tmp;
	u_int i;

	cell[4] = 1;
	cell[5] = flags;

	tmp = cps2atmf(er);
	cell[6] = tmp >> 8;
	cell[7] = tmp;

	tmp = cps2atmf(ccr);
	cell[8] = tmp >> 8;
	cell[9] = tmp;

	tmp = cps2atmf(mcr);
	cell[10] = tmp >> 8;
	cell[11] = tmp;

	cell[12] = cell[13] = cell[14] = cell[15] = 0;
	cell[16] = cell[17] = cell[18] = cell[19] = 0;

	for (i = 20; i < 50; i++)
		cell[i] = 0x6A;

	/* compute CRC */
	cell[50] = 0;
	cell[51] = 0;
	tmp = gen_crc10(0, cell + 4, 48);
	cell[50] = tmp >> 8;
	cell[51] = tmp;
}

static void
rmb_send(void)
{
	u_char cell[52];

	cell_mkheader(cell, 6);
	rm_makecell(cell, 0x80 | (ci << 5) | (ni << 4) | (ng << 6),
	    new_er < 0 ? 0 : new_er, 0, 0);
	cell_send(cell);
}

static void
data_send(void)
{
	u_char cell[52];

	cell_mkheader(cell, 1);
	cell_send(cell);
}

static void
process_rm(u_char *cell, u_int clp)
{
	double er, ccr, mcr;
	uint8_t flags;

	if (verbose)
		rm_dump(cell);

	if (gen_crc10(0, cell + 4, 48) != 0) {
		printf("bad RM CRC-10\n");
		ucells.total++;
		return;
	}
	if (cell[4] != 1) {
		printf("RM: proto=%u\n", cell[5]);
		ucells.total++;
		return;
	}

	flags = cell[5];

	if (flags & 0x80) {
		/* backward cell */
		if (clp)
			brm1cells.total++;
		else
			brm0cells.total++;
		return;
	}
	if (clp)
		frm1cells.total++;
	else
		frm0cells.total++;

	/* decode rates */
	er = atmf2cr((cell[6] << 8) | cell[7]);
	ccr = atmf2cr((cell[8] << 8) | cell[9]);
	mcr = atmf2cr((cell[10] << 8) | cell[11]);

	if (er != last_er) {
		printf("Forward ER changed to %f (ccr=%f, mcr=%f)\n",
		    er, ccr, mcr);
		last_er = er;
	}
	if (stop)
		return;

	flags = 0x80 | (ni << 4) | (ci << 5) | (ng << 6);
	if (new_er >= 0)
		er = new_er;

	/* put together the backward cell */
	cell_mkheader(cell, 6);
	rm_makecell(cell, flags, er, ccr, mcr);
	cell_send(cell);
}

/*
 * Input from the data socket
 */
static void
input_data(void)
{
	u_char buf[1000];
	char hook[NG_HOOKLEN + 1];
	u_int vpi, vci, pti, clp;
	int ret;

	ret = NgRecvData(dsock, buf, sizeof(buf), hook);
	if (ret == -1)
		err(1, "NgRecvData");
	if (strcmp(hook, RXHOOK) != 0)
		return;

	/* extract cell header */
	vpi = (buf[0] << 4) | (buf[1] >> 4);
	vci = ((buf[1] & 0xf) << 12) | (buf[2] << 4) | (buf[3] >> 4);
	pti = (buf[3] >> 1) & 0x7;
	clp = buf[3] & 1;

	if (pti == 6) {
		process_rm(buf, clp);
	} else if (pti == 0 || pti == 1) {
		/* data cell */
		dcells.total++;
	} else
		ucells.total++;
}

static void
input_user(void)
{
	char buf[LINE_MAX], *end, *ptr;
	size_t len;
	double tmp;

	if (fgets(buf, LINE_MAX, stdin) == NULL)
		return;
	if ((len = strlen(buf)) == 0)
		return;

	if (buf[len - 1] == '\n')
		buf[len - 1] = '\0';

	if (buf[0] == '\0') {
		printf("NI=%u CI=%u NG=%u ER=%f %sSTOPPED\n", ni, ci, ng,
		    new_er, stop ? "" : "UN");
		return;
	}

	ptr = buf;
	while (*ptr != '\0') {
		if (*ptr == 'n') {
			ni = !ni;
			printf("NI set to %u\n", ni);
			ptr++;
		} else if (*ptr == 'c') {
			ci = !ci;
			printf("CI set to %u\n", ci);
			ptr++;
		} else if (*ptr == 'x') {
			new_er = -1;
			printf("ER reset\n");
			ptr++;
		} else if (*ptr == 'g') {
			ng = !ng;
			printf("NG set to %u\n", ng);
			ptr++;
		} else if (*ptr == 'b') {
			rmb_send();
			printf("Cell sent\n");
			ptr++;
		} else if (*ptr == 's') {
			stop = !stop;
			printf(stop ? "stopped\n" : "unstopped\n");
			ptr++;
		} else if (*ptr == 'd') {
			data_send();
			printf("Cell sent\n");
			ptr++;
		} else {
			tmp = strtod(ptr, &end);
			if (end != ptr) {
				new_er = tmp;
				ptr = end;
			} else {
				printf("???%c???\n", *ptr++);
			}
		}
	}
}

static void
attach(const char *dev)
{
	struct ngm_connect cmsg;
	struct kevent change[4];

	NgSetDebug(0);
	if (NgMkSockNode(NULL, &csock, &dsock))
		err(1, "NgMkSockNode");

	/* connect the transmit hook */
	snprintf(cmsg.path, sizeof(cmsg.path), "%s:", dev);
	snprintf(cmsg.ourhook, sizeof(cmsg.ourhook), TXHOOK);
	snprintf(cmsg.peerhook, sizeof(cmsg.peerhook), "txabr%lu",
	    (u_long)getpid());

	if (NgSendMsg(csock, ".", NGM_GENERIC_COOKIE, NGM_CONNECT,
	    &cmsg, sizeof(cmsg)) == -1)
		err(1, "connect(%s,%s,%s)", cmsg.path, cmsg.ourhook,
		    cmsg.peerhook);

	strcpy(tparam.name, cmsg.peerhook);

	if (NgSendMsg(csock, cmsg.path, NGM_ATM_COOKIE, NGM_ATM_CPCS_INIT,
	    &tparam, sizeof(tparam)) == -1)
		err(1, "cpcs_init");

	/* connect the receive hook */
	snprintf(cmsg.ourhook, sizeof(cmsg.ourhook), RXHOOK);
	snprintf(cmsg.peerhook, sizeof(cmsg.peerhook), "rxabr%lu",
	    (u_long)getpid());

	if (NgSendMsg(csock, ".", NGM_GENERIC_COOKIE, NGM_CONNECT,
	    &cmsg, sizeof(cmsg)) == -1)
		err(1, "connect(%s,%s,%s)", cmsg.path, cmsg.ourhook,
		    cmsg.peerhook);

	strcpy(rparam.name, cmsg.peerhook);

	if (NgSendMsg(csock, cmsg.path, NGM_ATM_COOKIE, NGM_ATM_CPCS_INIT,
	    &rparam, sizeof(rparam)) == -1)
		err(1, "cpcs_init");

	fprintf(stderr, "cpcs attached\n");

	kq = kqueue();
	if (kq == -1)
		err(1, "kqueue");

	EV_SET(&change[0], dsock, EVFILT_READ, EV_ADD,
	    NOTE_LOWAT, 50, input_data);
	EV_SET(&change[1], fileno(stdin), EVFILT_READ, EV_ADD,
	    0, 0, input_user);
	EV_SET(&change[2], SIGINFO, EVFILT_SIGNAL, EV_ADD,
	    0, 0, print_info);
	EV_SET(&change[3], 1, EVFILT_TIMER, EV_ADD,
	    100, 0, process_timer);
	if (kevent(kq, change, 3, NULL, 0, NULL) == -1)
		err(1, "kevent(change)");
}

int
main(int argc, char *argv[])
{
	int opt;
	u_long res;
	char *end;
	const char *dev = "hatm0";
	int ret, i;
	struct kevent ev[10];
	struct timeval tv;

	while ((opt = getopt(argc, argv, "c:d:hsv")) != -1)
		switch (opt) {

		  case 'c':
			res = strtoul(optarg, &end, 10);
			if (*end == '.') {
				if (res > 0xff)
					errx(1, "vpi too large: %lu", res);
				rparam.vpi = tparam.vpi = res;
				end++;
				res = strtoul(end, &end, 10);
			} else
				tparam.vpi = 0;
			if (*end != '\0')
				errx(1, "bad [vpi.]vci argument '%s'", optarg);
			if (res > 0xffff)
				errx(1, "vci too large: %lu", res);
			tparam.vci = res;
			rparam.vci = res + 1;
			break;

		  case 'd':
			dev = optarg;
			break;

		  case 'h':
			fprintf(stderr, usgtxt);
			return (0);

		  case 's':
			stop = 1;
			break;

		  case 'v':
			verbose++;
			break;
		}

	setvbuf(stdout, NULL, _IOLBF, 0);

	gen_crc10_tab();
	attach(dev);

	gettimeofday(&tv, NULL);
	t_start = t_last = tv.tv_sec * 1000000ULL + tv.tv_usec;

	for (;;) {
		ret = kevent(kq, NULL, 0, ev, 10, NULL);
		if (ret == -1)
			err(1, "kevent");
		if (ret == 0) {
			printf("timeout\n");
			continue;
		}
		for (i = 0; i < ret; i++)
			((void (*)(void))ev[i].udata)();
	}

	return (0);
}
