/*
 * Copyright (c) 2001-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * $Id: atmrcv.c,v 1.3 2003/07/15 11:39:12 catscvs Exp $
 *
 * Simple program to receive via ng_atm
 */
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/endian.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <err.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <netgraph.h>
#include <netgraph/atm/ng_atm.h>
#include <net/if_atm.h>
#include <arpa/inet.h>
#include <math.h>

#define MAXMSG 65536

/* netgraph control/data socket */
static int csock, dsock;

/* verbosity */
static int verbose;

/* socket buffer */
static int sockbuf;

/* help text */
static const char usgtxt[] = "\
ATM traffic receiver.\n\
usage: atmrcv [-vhm] [-b sockbuf] [-c [vpi.]vci] [-d device] [-t osecs]\n\
options:\n\
  -b	set socket buffer size\n\
  -c	specify channel (default 0.66)\n\
  -d	specify device (default hatm0)\n\
  -H	delay histogram\n\
  -h	print this info\n\
  -t	specify the number of seconds between outputs\n\
  -m	only bandwidth measurement\n\
  -v	be verbose\n\
";

int bwonly;
int doinfo;
int dooutput;

uint64_t	delay_cnt;
double		delay_sum;
double		delay_qsum;

uint64_t	old_delay_cnt;
double		old_delay_sum;
double		old_delay_qsum;

uint64_t	lost, recvd;
uint64_t	old_lost, old_recvd;

uint64_t	cells_sum;
uint64_t	old_cells_sum;
double		old_secs;

#define DELAY_HIST_MAX	40000	/* usec */
#define DELAY_HIST_STEP	1000	/* usec */
uint64_t	delay_hist[40];
int		do_delay_hist;

struct timeval	pstart;

struct data {
	uint64_t count;
	uint32_t sec;
	uint32_t usec;
	uint64_t bcount;
};

static void
on_info(int s __unused)
{
	doinfo = 1;
}

static void
on_alrm(int s __unused)
{
	dooutput = 1;
}

static void
start_recv(const char *dev, u_int vpi, u_int vci)
{
	struct ngm_connect conn;
	struct ngm_atm_cpcs_init cpcs_init;

	NgSetDebug(0);
	if (NgMkSockNode(NULL, &csock, &dsock))
		err(1, "NgMkSockNode");

	if (sockbuf != 0 && setsockopt(dsock, SOL_SOCKET, SO_RCVBUF,
	    &sockbuf, sizeof(sockbuf)) == -1)
		err(1, "SO_RCVBUF %d", sockbuf);

	snprintf(conn.path, sizeof(conn.path), "%s:", dev);
	snprintf(conn.ourhook, sizeof(conn.ourhook), "recv");
	snprintf(conn.peerhook, sizeof(conn.peerhook), "atmrcv%lu",
	    (u_long)getpid());

	if (NgSendMsg(csock, ".", NGM_GENERIC_COOKIE, NGM_CONNECT,
	    &conn, sizeof(conn)) == -1)
		err(1, "connect(%s,%s,%s)", conn.path, conn.ourhook,
		    conn.peerhook);

	memset(&cpcs_init, 0, sizeof(cpcs_init));
	strcpy(cpcs_init.name, conn.peerhook);
	cpcs_init.flags = ATMIO_FLAG_NOTX;
	cpcs_init.vci = vci;
	cpcs_init.vpi = vpi;
	cpcs_init.aal = ATMIO_AAL_5;
	cpcs_init.pcr = 0;
	cpcs_init.traffic = ATMIO_TRAFFIC_UBR;

	if (NgSendMsg(csock, conn.path, NGM_ATM_COOKIE, NGM_ATM_CPCS_INIT,
	    &cpcs_init, sizeof(cpcs_init)) == -1)
		err(1, "cpcs_init");

	fprintf(stderr, "cpcs attached\n");
}

static void
print_info(void)
{
	double secs;
	struct timeval now;
	double mean, std;
	u_int i;

	gettimeofday(&now, NULL);
	timersub(&now, &pstart, &now);
	secs = now.tv_sec + now.tv_usec / 1000000.0;

	if (!bwonly) {
		if (delay_cnt > 0)
			mean = delay_sum / delay_cnt;
		else
			mean = 0;
		if (delay_cnt > 1)
			std = sqrt((delay_qsum - mean * delay_sum) /
			    (delay_cnt - 1));
		else
			std = 0;
		printf("%llu packets, delay=%f/%f; \n",
		    (unsigned long long)delay_cnt, mean, std);
	}

	printf("%llu cells; %f cps\n", (unsigned long long)cells_sum,
	    cells_sum / secs);

	if (!bwonly && do_delay_hist) {
		for (i = 0; i < DELAY_HIST_MAX / DELAY_HIST_STEP; i++)
			if (do_delay_hist < 2 || delay_hist[i] > 0)
				printf("%u %llu\n", i * DELAY_HIST_STEP,
				    (unsigned long long)delay_hist[i]);
	}
}

static void
print_output(void)
{
	double mean, std;
	uint64_t cnt;
	double sum, qsum, secs;
	struct timeval now;
	uint64_t r, l, c;
	double proc;

	gettimeofday(&now, NULL);
	timersub(&now, &pstart, &now);
	secs = now.tv_sec + now.tv_usec / 1000000.0;

	if (!bwonly) {

		/* compute delay since last output */
		cnt = delay_cnt - old_delay_cnt;
		sum = delay_sum - old_delay_sum;
		qsum = delay_qsum - old_delay_qsum;

		old_delay_cnt = delay_cnt;
		old_delay_sum = delay_sum;
		old_delay_qsum = delay_qsum;

		mean = std = 0;
		if (cnt > 0)
			mean = sum / cnt;

		if (cnt > 1)
			std = sqrt((qsum - mean * sum) / (cnt - 1));

		printf("%.1f %f %f", secs, mean, std);

		/* compute losses since last output */
		r = recvd - old_recvd;
		l = lost - old_lost;

		old_recvd = recvd;
		old_lost = lost;

		if ((r + l) == 0)
			proc = 0.0;
		else
			proc = (100.0 * l) / (r + l);
		printf(" %llu %llu %.1f", (unsigned long long)r,
		    (unsigned long long)l, proc);
	}

	/* compute cell rate since last output */
	c = cells_sum - old_cells_sum;
	old_cells_sum = cells_sum;

	printf(" %llu %.1f", (unsigned long long)c, c / (secs - old_secs));
	old_secs = secs;

	printf("\n");
	fflush(stdout);
}

static void
dump_buf(const u_char *buf, size_t len)
{
	size_t i;

	fprintf(stderr, "len=%zu:", len);
	for (i = 0; i < len; i++)
		fprintf(stderr, " %02x", *buf++);
	fprintf(stderr, "\n");
}

static void
start_timer(u_int out_interval)
{
	struct sigaction sa;
	struct itimerval itv;

	sa.sa_flags = 0;
	sigemptyset(&sa.sa_mask);
	sa.sa_handler = on_alrm;

	if (sigaction(SIGALRM, &sa, NULL) == -1)
		err(1, "sigaction");

	itv.it_interval.tv_sec = out_interval;
	itv.it_interval.tv_usec = 0;
	itv.it_value = itv.it_interval;

	if (setitimer(ITIMER_REAL, &itv, NULL) == -1)
		err(1, "setitimer");
}

int
main(int argc, char *argv[])
{
	int opt;
	u_long res;
	char *end;
	u_int vpi = 0, vci = 66;
	const char *dev = "hatm0";
	char *buf;
	ssize_t ret;
	struct sigaction sa;
	struct timeval now, diff, snd;
	double usec;
	uint64_t pktno, last_pktno;
	uint64_t next, bcount;
	u_long out_interval = 0;
	u_int idx;
	int trunning = 0;
	struct data *data;

	while ((opt = getopt(argc, argv, "b:c:d:Hhmt:v")) != -1)
		switch (opt) {

		  case 'b':
			errno = 0;
			res = strtol(optarg, &end, 10);
			if (*end != '\0' || errno != 0 || res <= 0)
				errx(1, "bad socket buffer size");
			sockbuf = res;
			break;

		  case 'c':
			res = strtoul(optarg, &end, 10);
			if (*end == '.') {
				if (res > 0xff)
					err(1, "vpi too large: %lu", res);
				vpi = res;
				end++;
				res = strtoul(end, &end, 10);
			} else
				vpi = 0;
			if (*end != '\0')
				errx(1, "bad [vpi.]vci argument '%s'", optarg);
			if (res > 0xffff)
				err(1, "vci too large: %lu", res);
			vci = res;
			break;

		  case 'd':
			dev = optarg;
			break;

		  case 'H':
			do_delay_hist++;
			break;

		  case 'h':
			fprintf(stderr, "%s", usgtxt);
			return (0);

		  case 'm':
			bwonly = 1;
			break;

		  case 't':
			out_interval = strtoul(optarg, &end, 10);
			if (*end != '\0')
				err(1, "bad -o argument '%s'", optarg);
			break;

		  case 'v':
			verbose++;
			break;
		}

	argc -= optind;
	argv += optind;

	if (argc != 0)
		errx(1, "extra args");

	start_recv(dev, vpi, vci);

	if ((buf = malloc(MAXMSG)) == NULL)
		err(1, NULL);
	memset(buf, 0, MAXMSG);
	data = (struct data *)(void *)buf;

	sa.sa_flags = 0;
	sigemptyset(&sa.sa_mask);
	sa.sa_handler = on_info;

	if (sigaction(SIGINFO, &sa, NULL) == -1)
		err(1, "sigaction");

	setvbuf(stdout, NULL, _IOLBF, 0);

	gettimeofday(&pstart, NULL);
	next = 0;
	pktno = 0;
	for (;;) {
		ret = NgRecvData(dsock, buf, MAXMSG, NULL);
		if (ret == -1) {
			if (errno == EINTR) {
				if (doinfo) {
					print_info();
					doinfo = 0;
				}
				if (dooutput) {
					print_output();
					dooutput = 0;
				}
				continue;
			}
			err(1, "NgRecvData");
		}
		gettimeofday(&now, NULL);
		if (out_interval != 0 && !trunning) {
			start_timer(out_interval);
			trunning = 1;
		}
		cells_sum += (ret + 8 + 47) / 48;

		if (!bwonly) {
			bcount = next;
			if (ret >= 24) {
				last_pktno = pktno;
				pktno = le64toh(data->count);
				snd.tv_sec = le32toh(data->sec);
				snd.tv_usec = le32toh(data->usec);
				bcount = le64toh(data->bcount);

				timersub(&now, &snd, &diff);
				usec = 1000000.0 * diff.tv_sec + diff.tv_usec;
				delay_cnt++;
				delay_sum += usec;
				delay_qsum += usec * usec;

				if (bcount < next) {
					/* ignore that one */
					if (verbose)
						dump_buf(buf, ret);
					warnx("bcount %llu/%llu <- %llu/%llu",
					    (unsigned long long)last_pktno,
					    (unsigned long long)next,
					    (unsigned long long)pktno,
					    (unsigned long long)bcount);
				} else {
					if (bcount > next) {
						if (verbose)
							warnx("bcount %llu/%llu"
							    " -> %llu/%llu",
							    (unsigned long long)last_pktno,
							    (unsigned long long)next,
							    (unsigned long long)pktno,
							    (unsigned long long)bcount);
						lost += bcount - next;
					}
					recvd++;
					next = bcount + 1;
				}

				if (do_delay_hist && usec < DELAY_HIST_MAX) {
					idx = (usec + DELAY_HIST_STEP / 2)
					    / DELAY_HIST_STEP;
					if (idx < DELAY_HIST_MAX / DELAY_HIST_STEP)
						delay_hist[idx]++;
				}
				if (verbose > 1) {
					printf("packet %zu bytes, %llu %.0f\n",
					    ret, (unsigned long long)pktno,
					    usec);
				}
			} else {
				warnx("small packet");
				if (verbose > 1) {
					printf("packet %zu bytes, %ld.%06ld\n",
					    ret, snd.tv_sec, snd.tv_usec);
				}
			}
		}
		if (doinfo) {
			print_info();
			doinfo = 0;
		}
		if (dooutput) {
			print_output();
			dooutput = 0;
		}
	}

	if (verbose)
		print_info();

	return (0);
}
