/*
 * Copyright (c) 2001-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: atmsnd.c,v 1.4 2003/07/21 07:37:08 catscvs Exp $
 *
 * Author: Harti Brandt <harti@freebsd.org>
 *
 * Simple program to send via ng_atm
 *
 * This program sends bursts of packets. One can specify the size of each
 * packet (-m), the number of packets in each burst (-b), the time to
 * wait between bursts in milliseconds (-s) and the total number of bursts
 * to send (-n). Each packet contains the burst number, the packet number and
 * the local time. The program atmrcv can use these to compute loss rates,
 * delays and other parameters.
 */
#include <sys/time.h>
#include <sys/endian.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <err.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <netgraph.h>
#include <netgraph/atm/ng_atm.h>
#include <net/if_atm.h>
#include <arpa/inet.h>

/* netgraph control/data socket */
static int csock, dsock;

/* peek cell rate */
static struct ngm_atm_cpcs_init tparam = {
	vci:	66,
	rmtu:	10000,
	tmtu:	10000,
	aal:	ATMIO_AAL_5,
	traffic:ATMIO_TRAFFIC_CBR,
	pcr:	100
};
/* starting time of loop */
struct timeval proc_start;

/* verbosity */
static int verbose;

/* data sent */
uint64_t bytes_sent;
uint64_t cells_sent;
uint64_t mesg_sent;
uint64_t bursts_sent;

/* help text */
static const char usgtxt[] = "\
ATM traffic generator.\n\
usage: atmsnd [-vh] [-A aal] [-B mbs] [-b pkts] [-C cdf] [-c [vpi.[vci]]\n\
              [-D rdf] [-d device] [-E tbe] [-F rif] [-M mcr] [-m msgsiz]\n\
	      [-I icr] [-N nrm] [-n bursts] [-p pcr] [-S scr] [-s pps]\n\
	      [-T trm] [-t servcat]\n\
options:\n\
  -A	specify aal to use (0, 3, 4, 5 or raw)\n\
  -B	specify maximum burst size in cells for VBR\n\
  -b	specify the number of packets in each burst\n\
  -C	specify CDF for ABR\n\
  -c	specify channel (default 0.66)\n\
  -D	specify ABR rate decrease factor\n\
  -d	specify device (default hatm0)\n\
  -E	specify ABR TBE\n\
  -F	specify ABR rate increase factor\n\
  -h	print this info\n\
  -I	specify initial cell rate for ABR\n\
  -M	specify minimum cell rate\n\
  -m	size of each packet in bytes (default 1000)\n\
  -N	specify ABR NRM\n\
  -n	total number of bursts to send (default unlimited)\n\
  -p	peak cell rate\n\
  -S	specify sustainable cell rate for VBR\n\
  -s	milliseconds between bursts (default 1000)\n\
  -T	specify ABR TRM\n\
  -t	service category (default cbr)\n\
  -v	be verbose\n\
";

struct data {
	uint64_t count;
	uint32_t sec;
	uint32_t usec;
	uint64_t bcount;
};

int doinfo;

static void
on_info(int s __unused)
{
	doinfo = 1;
}

static void
start_send(const char *dev)
{
	struct ngm_connect connect;

	NgSetDebug(0);
	if (NgMkSockNode(NULL, &csock, &dsock))
		err(1, "NgMkSockNode");

	snprintf(connect.path, sizeof(connect.path), "%s:", dev);
	snprintf(connect.ourhook, sizeof(connect.ourhook), "send");
	snprintf(connect.peerhook, sizeof(connect.peerhook), "atmsnd%lu",
	    (u_long)getpid());

	if (NgSendMsg(csock, ".", NGM_GENERIC_COOKIE, NGM_CONNECT,
	    &connect, sizeof(connect)) == -1)
		err(1, "connect(%s,%s,%s)", connect.path, connect.ourhook,
		    connect.peerhook);

	strcpy(tparam.name, connect.peerhook);
	tparam.flags = 0;

	if (NgSendMsg(csock, connect.path, NGM_ATM_COOKIE, NGM_ATM_CPCS_INIT,
	    &tparam, sizeof(tparam)) == -1)
		err(1, "cpcs_init");

	fprintf(stderr, "cpcs attached\n");
}

static void
print_info(void)
{
	double secs;
	struct timeval end, diff;

	gettimeofday(&end, NULL);
	timersub(&end, &proc_start, &diff);
	secs = diff.tv_sec + diff.tv_usec / 1000000.0;

	printf("%f secs; %llu msgs; %f cells/sec; %f bytes/sec; %f msg/sec\n",
	    secs, (unsigned long long)mesg_sent, cells_sent / secs,
	    bytes_sent / secs, mesg_sent / secs);
}

static u_int
parse(const char *name, u_long max)
{
	u_long res;
	char *end;

	res = strtoul(optarg, &end, 10);
	if (*end != '\0')
		err(1, "bad %s '%s'", name, optarg);
	if (res > max)
		err(1, "%s too large: %lu", name, res);
	return (res);
}

static u_int
parse24(const char *name)
{
	return (parse(name, 0xffffff));
}

int
main(int argc, char *argv[])
{
	int opt;
	u_long res;
	char *end;
	const char *dev = "hatm0";
	u_int bytes_per_msg = 1000;
	u_int msec_burst = 1000;
	u_int cells;
	char *buf;
	ssize_t ret;
	struct timeval endt, now;
	struct timespec ts;
	struct sigaction sa;
	uint64_t count;
	uint64_t bcount;
	uint64_t bursts = 0;
	uint64_t step, ndiff;
	uint64_t nsecs;
	u_int burst = 1, b;
	struct data *data;

#if 0
	/* not settable yet */
	uint16_t	rmtu;		/* Receive maximum CPCS size */
	uint16_t	tmtu;		/* Transmit maximum CPCS size */
	uint16_t	adtf;		/* ABR: ADTF */
#endif

	while ((opt = getopt(argc, argv,
	    "A:B:b:C:c:D:d:E:F:hI:M:m:N:n:p:S:s:T:t:v")) != -1)
		switch (opt) {

		  case 'A':
			if (strcmp(optarg, "0") == 0)
				tparam.aal = ATMIO_AAL_0;
			else if (strcmp(optarg, "3") == 0)
				tparam.aal = ATMIO_AAL_34;
			else if (strcmp(optarg, "4") == 0)
				tparam.aal = ATMIO_AAL_34;
			else if (strcmp(optarg, "5") == 0)
				tparam.aal = ATMIO_AAL_5;
			else if (strcmp(optarg, "raw") == 0)
				tparam.aal = ATMIO_AAL_RAW;
			else
				errx(1, "bad AAL '%s'", optarg);
			break;

		  case 'B':
			tparam.mbs = parse24("mbs");
			break;

		  case 'b':
			res = strtoul(optarg, &end, 10);
			if (*end != '\0')
				errx(1, "bad pkts per burst '%s'", optarg);
			if ((burst = res) == 0)
				errx(1, "cannot have 0 packets per burst");
			break;

		  case 'C':
			tparam.cdf = parse("cdf", 0x7);
			break;

		  case 'c':
			res = strtoul(optarg, &end, 10);
			if (*end == '.') {
				if (res > 0xff)
					errx(1, "vpi too large: %lu", res);
				tparam.vpi = res;
				end++;
				res = strtoul(end, &end, 10);
			} else
				tparam.vpi = 0;
			if (*end != '\0')
				errx(1, "bad [vpi.]vci argument '%s'", optarg);
			if (res > 0xffff)
				errx(1, "vci too large: %lu", res);
			tparam.vci = res;
			break;

		  case 'D':
			tparam.rdf = parse("rdf", 0xf);
			break;

		  case 'd':
			dev = optarg;
			break;

		  case 'E':
			tparam.tbe = parse24("tbe");
			break;

		  case 'F':
			tparam.rif = parse("rif", 0xf);
			break;

		  case 'h':
			fprintf(stderr, "%s", usgtxt);
			return (0);

		  case 'I':
			tparam.icr = parse24("icr");
			break;

		  case 'M':
			tparam.mcr = parse24("mcr");
			break;

		  case 'm':
			res = strtoul(optarg, &end, 10);
			if (*end != '\0')
				errx(1, "bad bytes_per_msg '%s'", optarg);
			bytes_per_msg = res;
			break;

		  case 'N':
			tparam.nrm = parse("nrm", 0xf);
			break;

		  case 'n':
			bursts = strtoull(optarg, &end, 10);
			if (*end != '\0')
				errx(1, "bad no of bursts '%s'", optarg);
			break;

		  case 'p':
			tparam.pcr = parse24("pcr");
			break;

		  case 'S':
			tparam.scr = parse24("scr");
			break;


		  case 's':
			res = strtoul(optarg, &end, 10);
			if (*end != '\0')
				errx(1, "bad msec_burst '%s'", optarg);
			msec_burst = res;
			break;

		  case 'T':
			tparam.trm = parse("trm", 0xf);
			break;

		  case 't':
			if (strcmp(optarg, "cbr") == 0)
				tparam.traffic = ATMIO_TRAFFIC_CBR;
			else if (strcmp(optarg, "abr") == 0)
				tparam.traffic = ATMIO_TRAFFIC_ABR;
			else if (strcmp(optarg, "vbr") == 0)
				tparam.traffic = ATMIO_TRAFFIC_VBR;
			else if (strcmp(optarg, "ubr") == 0)
				tparam.traffic = ATMIO_TRAFFIC_UBR;
			else
				errx(1, "bad traffic '%s'", optarg);
			break;

		  case 'v':
			verbose = 1;
			break;
		}

	argc -= optind;
	argv += optind;

	if (argc != 0)
		errx(1, "extra args");

	if (bytes_per_msg < 30) {
		warnx("adjusting bytes/msg to 30");
		bytes_per_msg = 30;
	}

	start_send(dev);

	nsecs = 1000000ULL * msec_burst;

	if ((buf = malloc(bytes_per_msg)) == NULL)
		err(1, NULL);
	memset(buf, 0, bytes_per_msg);
	data = (struct data *)(void *)buf;

	sa.sa_flags = SA_RESTART;
	sigemptyset(&sa.sa_mask);
	sa.sa_handler = on_info;

	if (sigaction(SIGINFO, &sa, NULL) == -1)
		err(1, "sigaction");

	count = 0;
	bcount = 0;
	gettimeofday(&proc_start, NULL);
	step = 0;
	for (;;) {
		data->count = htole64(count++);
		for (b = 0; b < burst; b++) {
			gettimeofday(&now, NULL);

			data->bcount = htole64(bcount++);
			data->sec = htole32(now.tv_sec);
			data->usec = htole32(now.tv_usec);

			ret = NgSendData(dsock, "send", buf, bytes_per_msg);
			if (ret == -1)
				err(1, "sending");

			cells = (bytes_per_msg + 8 + 47) / 48;
			bytes_sent += cells * 48;;
			cells_sent += cells;
			mesg_sent++;
		}
		bursts_sent++;

		if (doinfo) {
			print_info();
			doinfo = 0;
		}

		if (bursts != 0 && bursts_sent == bursts)
			break;

		gettimeofday(&endt, NULL);
		timersub(&endt, &proc_start, &endt);
		ndiff = 1000000000LLU * endt.tv_sec + 1000 * endt.tv_usec;
		step += nsecs;

		if (ndiff >= step || (ndiff = step - ndiff) < 1000000) {
			if (verbose)
				printf("late %llu\n",
				    (unsigned long long)mesg_sent);
			continue;
		}

		ts.tv_sec = ndiff / 1000000000;
		ts.tv_nsec = ndiff % 1000000000;

		for (;;) {
			if (nanosleep(&ts, &ts) != -1)
				break;
			if (errno != EINTR)
				err(1, "nanosleep");
			if (ts.tv_sec == 0 && ts.tv_nsec == 0)
				break;
		}
	}

	if (verbose)
		print_info();

	/* wait for the output to drain */
	sleep(4);


	return (0);
}
